(* Interpreter -------------------------------------------------------------- *)

type state = var -> int64

let rec interpret_exp (s:state) (e:exp) : int64 =
  match e with
  | Var x -> s x
  | Add (e1, e2) -> Int64.add (interpret_exp s e1) (interpret_exp s e2)
  | Mul (e1, e2) -> Int64.mul (interpret_exp s e1) (interpret_exp s e2)
  | Const n -> n
  | Neg e -> Int64.neg (interpret_exp s e)

let update (s:state) (x:var) (v:int64) =
  fun y -> if x = y then v else s y

exception Halt of int64

let rec interpret_cmd (s:state) (c:cmd) : state =
  match c with
  | Skip -> s
  | Assn (x, e1) ->
    let v = interpret_exp s e1 in
    update s x v
  | IfLZ (e1, c1, c2) ->
    if Int64.compare (interpret_exp s e1) Int64.zero < 0
    then interpret_cmd s c1
    else interpret_cmd s c2
  | WhileLZ (e, c) ->
    if Int64.compare (interpret_exp s e) Int64.zero < 0 then
      interpret_cmd s (Seq(c, WhileLZ (e, c)))
    else
      s
  | Seq (c1, c2) ->
    let s1 = interpret_cmd s c1 in
    interpret_cmd s1 c2
  | Return e ->
    raise (Halt (interpret_exp s e))
  
let init_state : state = fun _ -> 0L

let answer program =
  try
    ignore (interpret_cmd init_state program "");
    0L;
  with Halt x -> x 
